from humpday.comparison.naming import M  #Silly https://www.linkedin.com/posts/tedpetrou_pythonquiz-activity-6766019617098924033-6jIL

from humpday.optimizers.alloptimizers import OPTIMIZERS
from humpday.objectives.allobjectives import OBJECTIVES
from humpday.objectives.allobjectives import CLASSIC_OBJECTIVES
from humpday.comparison.suggestions import suggest, recommend
from humpday.optimizers.meta import minimize


